@@ -20,7 +20,7 @@
 # along with Smart Package Manager; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 #
-from smart.backends.rpm.rpmver import splitarch
+from smart.backends.rpm.rpmver import checkver, splitarch
 from smart.util.strtools import globdistance
 from smart.cache import Loader, PackageInfo
 from smart.channel import FileChannel
@@ -30,6 +30,8 @@ from smart import *
 import locale
 import stat
 import os
+from datetime import datetime
+import time
 
 try:
     import rpmhelper
@@ -51,6 +53,26 @@ CRPMTAG_UPDATE_URL        = 1000023
 
 ENCODINGS = ["utf8", "iso-8859-1"]
 
+
+def get_header_filenames(header):
+    filenames = header[rpm.RPMTAG_OLDFILENAMES]
+    if not filenames and header[rpm.RPMTAG_BASENAMES]:
+        dirindexes = header[rpm.RPMTAG_DIRINDEXES]
+        if type(dirindexes) != list:
+            dirindexes = [dirindexes]
+        dirnames = header[rpm.RPMTAG_DIRNAMES]
+        if type(dirnames) != list:
+            dirnames = [dirnames]
+        basenames = header[rpm.RPMTAG_BASENAMES]
+        if type(basenames) != list:
+            basenames = [basenames]
+        filenames = [dirnames[dirindexes[i]] + basename
+                     for i, basename in enumerate(basenames)]
+    elif type(filenames) != list:
+        filenames = [filenames]
+    return filenames
+
+
 class RPMHeaderPackageInfo(PackageInfo):
 
     class LazyHeader(object):
@@ -64,6 +86,7 @@ class RPMHeaderPackageInfo(PackageInfo):
         PackageInfo.__init__(self, package, order)
         self._loader = loader
         self._path = None
+        self._change = None
 
     def getReferenceURLs(self):
         url = self._h[rpm.RPMTAG_URL]
@@ -83,6 +106,9 @@ class RPMHeaderPackageInfo(PackageInfo):
     def getMD5(self, url):
         return self._loader.getMD5(self)
 
+    def getBuildTime(self):
+        return self._h[rpm.RPMTAG_BUILDTIME]
+
     def getInstalledSize(self):
         return self._h[rpm.RPMTAG_SIZE]
 
@@ -100,12 +126,33 @@ class RPMHeaderPackageInfo(PackageInfo):
                 break
         return result
 
+    def _getHeaderArray(self, tag):
+        result = self._h and self._h[tag] or []
+        if result:
+            if type(result) != list:
+                result = [result]
+            for i in range(len(result)):
+                if type(result[i]) == str:
+                    for encoding in ENCODINGS:
+                        try:
+                            result[i] = result[i].decode(encoding)
+                        except UnicodeDecodeError:
+                            continue
+                        break
+        return result
+
     def getDescription(self):
         return self._getHeaderString(rpm.RPMTAG_DESCRIPTION)
 
     def getSummary(self):
         return self._getHeaderString(rpm.RPMTAG_SUMMARY)
 
+    def getSource(self):
+        sourcerpm = self._getHeaderString(rpm.RPMTAG_SOURCERPM)
+        sourcerpm = sourcerpm.replace(".src", "")
+        sourcerpm = sourcerpm.replace(".nosrc", "")
+        return sourcerpm.replace(".rpm", "")
+    
     def getGroup(self):
         s = self._loader.getGroup(self._package)
         for encoding in ENCODINGS:
@@ -118,11 +165,24 @@ class RPMHeaderPackageInfo(PackageInfo):
             s = ""
         return s
 
+    def getLicense(self):
+        return self._getHeaderString(rpm.RPMTAG_LICENSE)
+
+    def getChangeLog(self):
+        if self._change is None:
+            logtime = self._getHeaderArray(rpm.RPMTAG_CHANGELOGTIME)
+            self._change = []
+            if len(logtime) > 0:
+                logname = self._getHeaderArray(rpm.RPMTAG_CHANGELOGNAME)
+                change = self._getHeaderArray(rpm.RPMTAG_CHANGELOGTEXT)
+                for i in range(len(change)):
+                    self._change.append(datetime.fromtimestamp(logtime[i]).strftime("%Y-%m-%d")+"  "+ logname[i])
+                    self._change.append("  " + change[i])
+        return self._change
+
     def getPathList(self):
         if self._path is None:
-            paths = self._h[rpm.RPMTAG_OLDFILENAMES]
-            if type(paths) != list:
-                paths = [paths]
+            paths = get_header_filenames(self._h)
             modes = self._h[rpm.RPMTAG_FILEMODES]
             if modes:
                 if type(modes) != list:
@@ -196,7 +256,7 @@ class RPMHeaderLoader(Loader):
             if h[1106]: # RPMTAG_SOURCEPACKAGE
                 continue
             arch = h[1022] # RPMTAG_ARCH
-            if rpm.archscore(arch) == 0:
+            if getArchScore(arch) == 0:
                 continue
 
             name = h[1000] # RPMTAG_NAME
@@ -218,7 +278,7 @@ class RPMHeaderLoader(Loader):
                     vi = v[i]
                     if vi and vi[:2] == "0:":
                         vi = vi[2:]
-                    if ni == name and vi == version:
+                    if ni == name and checkver(vi, version):
                         prvdict[(NPrv, intern(ni), versionarch)] = True
                     else:
                         prvdict[(Prv, intern(ni), vi or None)] = True
@@ -250,7 +310,7 @@ class RPMHeaderLoader(Loader):
                             # RPMSENSE_SCRIPT_POSTUN == 7744
                             reqdict[(f[i]&7744 and PreReq or Req,
                                      intern(ni), r, vi)] = True
-                reqargs = reqdict.keys()
+                reqargs = collapse_libc_requires(reqdict.keys())
             else:
                 reqargs = None
 
@@ -296,6 +356,14 @@ class RPMHeaderLoader(Loader):
             else:
                 upgargs = [obstup]
 
+            disttag = h[1155] # RPMTAG_DISTTAG
+            distepoch = h[1218] # RPMTAG_DISTEPOCH
+            if disttag:
+                distversion = "%s-%s" % (version, disttag)
+                if distepoch:
+                    distversion += distepoch
+                versionarch = "%s@%s" % (distversion, arch)
+
             pkg = self.buildPackage((Pkg, name, versionarch),
                                     prvargs, reqargs, upgargs, cnfargs)
             pkg.loaders[self] = offset
@@ -323,10 +391,8 @@ class RPMHeaderLoader(Loader):
                 searcher.addResult(pkg, ratio)
                 continue
             if searcher.path:
-                paths = h[rpm.RPMTAG_OLDFILENAMES]
+                paths = get_header_filenames(h)
                 if paths:
-                    if type(paths) != list:
-                        paths = [paths]
                     for spath, cutoff in searcher.path:
                         for path in paths:
                             _, newratio = globdistance(spath, path, cutoff, ic)
@@ -474,10 +540,7 @@ class RPMHeaderListLoader(RPMHeaderLoader):
         h, offset = rpm.readHeaderFromFD(file.fileno())
         bfp = self.buildFileProvides
         while h:
-            fnlst = h[1027] # RPMTAG_OLDFILENAMES
-            if type(fnlst) != list:
-                fnlst = [fnlst]
-            for fn in fnlst:
+            for fn in get_header_filenames(h):
                 fn = fndict.get(fn)
                 if fn and offset in self._offsets:
                     bfp(self._offsets[offset], (RPMProvides, fn, None))
@@ -487,10 +550,7 @@ class RPMHeaderListLoader(RPMHeaderLoader):
     def loadFileProvidesHDL(self, fndict):
         bfp = self.buildFileProvides
         for offset, h in enumerate(self._hdl):
-            fnlst = h[1027] # RPMTAG_OLDFILENAMES
-            if type(fnlst) != list:
-                fnlst = [fnlst]
-            for fn in fnlst:
+            for fn in get_header_filenames(h):
                 fn = fndict.get(fn)
                 if fn and offset in self._offsets:
                     bfp(self._offsets[offset], (RPMProvides, fn, None))
@@ -518,12 +578,18 @@ class URPMILoader(RPMHeaderListLoader):
     def __init__(self, filename, baseurl, listfile):
         RPMHeaderListLoader.__init__(self, filename, baseurl)
         self._prefix = {}
-        if listfile:
-            for entry in open(listfile):
-                if entry[:2] == "./":
-                    entry = entry[2:]
-                dirname, basename = os.path.split(entry.rstrip())
-                self._prefix[basename] = dirname
+        self._flagdict = None
+
+    def setErrataFlags(self, flagdict):
+        self._flagdict = flagdict
+    
+    def buildPackage(self, pkgargs, prvargs, reqargs, upgargs, cnfargs):
+        pkg = Loader.buildPackage(self, pkgargs, prvargs, reqargs, upgargs, cnfargs)
+        name = pkgargs[1]
+        if hasattr(self, '_flagdict') and self._flagdict and name in self._flagdict:
+            if sysconf.getReadOnly():
+                pkgconf.setFlag(self._flagdict[name], name, "=", pkgargs[2])
+        return pkg
 
     def getFileName(self, info):
         h = info._h
@@ -565,14 +631,18 @@ class RPMDBLoader(RPMHeaderLoader):
             prog.show()
         prog.add(left)
 
-    if rpmhelper:
-        def getHeader(self, pkg):
+    def getHeader(self, pkg):
+        if rpmhelper:
             mi = rpmhelper.dbMatch(getTS(), 0, pkg.loaders[self])
-            return mi.next()
-    else:
-        def getHeader(self, pkg):
+        else:
             mi = getTS().dbMatch(0, pkg.loaders[self])
+        try:
             return mi.next()
+        except StopIteration:
+            class NullHeader(object):
+                def __getitem__(self, key):
+                    return None
+            return NullHeader()
 
     def getURL(self):
         return None
@@ -677,10 +747,7 @@ class RPMDirLoader(RPMHeaderLoader):
                 iface.error("%s: %s" % (os.path.basename(filepath), e))
             else:
                 file.close()
-                fnlst = h[1027] # RPMTAG_OLDFILENAMES
-                if type(fnlst) != list:
-                    fnlst = [fnlst]
-                for fn in fnlst:
+                for fn in get_header_filenames(h):
                     fn = fndict.get(fn)
                     if fn:
                         bfp(self._offsets[i], (RPMProvides, fn, None))
